#include <linux/config.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/err.h>
#include <linux/devfs_fs_kernel.h>

#include "task_security.h"
#include "task_contexts.h"
#include "chrdev.h"

static int tcontext_new_open(struct inode *inode, struct file *file)
{
	int err;
	task_context_t *tc;
	
	tc = create_task_context(task_context(current));
	if (IS_ERR(tc)) { 
		printk(KERN_DEBUG "create_task_context() failed\n");
		return PTR_ERR(tc);
	}

	err = attach_task_context(current, tc);
	if (err) {
		printk(KERN_DEBUG "attach_task_context() failed\n");
		kobject_put(&tc->kobj);
		return err;
	}
	
	/* Since both create_new_context and attach_task_context increment
	 tc->kobj.refcount, we'd end up with 2 references if we
	 don't release one of them here. */
	kobject_put(&tc->kobj);
	
	return 0;
}


static int tcontext_disablenetwork_open(struct inode *inode, struct file *file)
{
	int err;
	
	if (!task_security(current)) {
		err = alloc_task_security(current);
		if (err)
			return err;
	}
	task_security(current)->disablenetwork_mode = 1;
	return 0;
}

static int tcontext_open(struct inode *inode, struct file *file)
{
	switch (iminor(inode)) {
	case TCONTEXT_MINOR_NEW:
		return tcontext_new_open(inode, file);
	case TCONTEXT_MINOR_DISABLENETWORK:
		return tcontext_disablenetwork_open(inode, file);
	default:
		return -ENODEV;
	}
}

static struct file_operations tcontext_fops = {
	open: tcontext_open,
	//release:
	//ioctl:
	owner: THIS_MODULE,
};

int tcontext_chrdev_register(void)
{
	if (register_chrdev(TCONTEXT_MAJOR, "tcontext", &tcontext_fops)) {
		printk(KERN_ERR "tcontext: failed to register character device\n");
		return -ENOMEM;
		
	}
	devfs_mk_dir("tcontext");
	devfs_mk_cdev(MKDEV(TCONTEXT_MAJOR, TCONTEXT_MINOR_NEW),
			S_IFCHR | S_IRUGO | S_IWUGO,
			"tcontext/new");
	devfs_mk_cdev(MKDEV(TCONTEXT_MAJOR, TCONTEXT_MINOR_DISABLENETWORK),
			S_IFCHR | S_IRUGO | S_IWUGO,
			"tcontext/disablenetwork");
	return 0;
}

void tcontext_chrdev_unregister(void)
{
	devfs_remove("tcontext/disablenetwork");
	devfs_remove("tcontext/new");
	devfs_remove("tcontext");
	unregister_chrdev(TCONTEXT_MAJOR, "tcontext");
}
