#include <linux/config.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/security.h>
#include "task_security.h"
#include "task_contexts.h"
#include "inode_security.h"
#include "sec_ops.h"

int ptrace_mode = TCONTEXT_PTRACE_NORMAL;

static int tcontext_task_alloc_security(struct task_struct *p)
{
	int err = 0;
	task_context_t *tc;
	/* Allocate the task_security_t structure */
	err = alloc_task_security(p);
	if (err)
		return err;
	
	/* Now, copy all the security attributes from current process to the
	 * new one. */
	if (!task_security(current))
		return 0;

	/* Attach the task context of the current process to the new process. */
	tc = task_context(current);
	if (tc) {
		err = attach_task_context(p, tc);
		if (err)
			return err;
	}

	/* disablenetwork() behaviour is inherited */
	if (task_security(current)) {
		task_security(p)->disablenetwork_mode = 
			task_security(current)->disablenetwork_mode;
	}
	return 0;
}

static void tcontext_task_free_security(struct task_struct *p)
{
	task_security_t *ts;
	if (task_context(p))
		detach_task_context(p);
	if (task_security(p)) {
		ts = task_security(p);
		task_security(p) = NULL;
		kfree(ts);
	}
}

static int tcontext_ptrace (struct task_struct * parent, struct task_struct * child)
{
	switch (ptrace_mode) {
	case TCONTEXT_PTRACE_NORMAL:
		if (is_task_subcontext(task_context(child),
				       task_context(parent)))
			return 0;
		return -EPERM;

	case TCONTEXT_PTRACE_DISABLE:
		return -EPERM;

	case TCONTEXT_PTRACE_ROOT:
		if (!task_context(parent) ||
		    task_context(parent) == root_task_context)
			return 0;
		return -EPERM;
	
	default:
		printk(KERN_ERR "BUG: unrecognized ptrace_mode %d set\n", ptrace_mode); 
		return -EPERM;
	}
}


static inline int tcontext_task_common_sec(struct task_struct *p)
{
	if (is_task_subcontext(task_context(p), task_context(current)))
		return 0;
	return -EPERM;
}

/* -------- Begin stub definitions -------- */

static int tcontext_task_setnice (struct task_struct *p, int nice)
{
	return tcontext_task_common_sec(p);
}

static int tcontext_task_setpgid (struct task_struct *p, int pgid)
{
	return tcontext_task_common_sec(p);
}

static int tcontext_task_setscheduler (struct task_struct *p, int policy, struct sched_param *lp)
{
	return tcontext_task_common_sec(p);
}

static int tcontext_task_kill (struct task_struct *p, struct siginfo *info, int sig)
{
	if (info && ((int) info == 1 || (int) info == 2 || SI_FROMKERNEL(info)))
		return 0;
	return tcontext_task_common_sec(p);
}

/* -------- End stub definitions -------- */


static int tcontext_inode_alloc_security(struct inode *inode)
{
	return alloc_inode_security(inode);
}

static void tcontext_inode_free_security(struct inode *inode)
{
	if (!inode_security(inode))
		return;
	if (i_task_context(inode))
		kobject_put(&i_task_context(inode)->kobj);
	
	free_inode_security(inode);
}

static void tcontext_task_to_inode(struct task_struct *p, struct inode *inode)
{
	if (!inode_security(inode) && alloc_inode_security(inode)) {
		printk(KERN_ERR "tcontext_task_to_inode: Unable to allocate inode->i_security for PID %d\n", p->pid);
		return;
	}
	inode_security(inode)->is_task_inode = 1;
	if (task_context(p)) {
		if (!kobject_get(&task_context(p)->kobj)) {
			printk(KERN_DEBUG "kobject_get failed.");					return;
		}
	}
	inode_security(inode)->context = task_context(p);
}

static int tcontext_inode_permission(struct inode *inode, int mask, struct nameidata *nd)
{
	if (!inode_security(inode) || !inode_security(inode)->is_task_inode)
		return 0;
	if (is_task_subcontext(i_task_context(inode), task_context(current)))
		return 0;
	return -EPERM;
}

static int tcontext_socket_bind (struct socket * sock,
			    struct sockaddr * address, int addrlen)
{
	if (task_security(current)
	    && task_security(current)->disablenetwork_mode
	    && ((struct sockaddr *)address)->sa_family != AF_UNIX)
		return -EACCES;
	return 0;
}

static int tcontext_socket_connect (struct socket * sock,
			       struct sockaddr * address, int addrlen)
{
	if (task_security(current)
	    && task_security(current)->disablenetwork_mode
	    && ((struct sockaddr *)address)->sa_family != AF_UNIX)
		return -EACCES;
	return 0;
}

static int tcontext_socket_create (int family, int type, int protocol)
{
	if (task_security(current)
	    && task_security(current)->disablenetwork_mode
	    && family != AF_UNIX)
		return -EACCES;
	return 0;
}

static int tcontext_socket_sendmsg (struct socket * sock,
			       struct msghdr * msg, int size)
{
	if (!task_security(current)
	    || !task_security(current)->disablenetwork_mode)
		return 0;
	
	/* sendto() a non-Unix socket */
	/* FIXME: Is this correct?  Is this safe? */
	if (msg->msg_namelen != 0 &&
	    ((struct sockaddr *)&msg->msg_name)->sa_family != AF_UNIX)
		return -EACCES;
	return 0;
}


struct security_operations tcontext_sec_ops = {
	ptrace: tcontext_ptrace,
	task_alloc_security: tcontext_task_alloc_security,
	task_free_security: tcontext_task_free_security,
	task_kill: tcontext_task_kill,
	task_setnice: tcontext_task_setnice,
	task_setscheduler: tcontext_task_setscheduler,
	task_getscheduler: tcontext_task_common_sec,
	task_setpgid: tcontext_task_setpgid,
	task_getpgid: tcontext_task_common_sec,
	task_getsid: tcontext_task_common_sec,
	inode_alloc_security: tcontext_inode_alloc_security,
	inode_free_security: tcontext_inode_free_security,
	task_to_inode: tcontext_task_to_inode,
	inode_permission: tcontext_inode_permission,
	socket_create: tcontext_socket_create,
	socket_bind: tcontext_socket_bind,
	socket_connect: tcontext_socket_connect,
	socket_sendmsg: tcontext_socket_sendmsg,
};


