#include <linux/config.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/err.h>

#include "task_security.h"
#include "task_contexts.h"

task_context_t *root_task_context = NULL;

static void task_context_release_kobj(struct kobject *kobj)
{
	task_context_t *tc = container_of(kobj, task_context_t, kobj);
	if (likely(tc != root_task_context)) {
		//kobject_get(kobj);
		kobject_unregister(kobj);
	}
	kfree(tc);
	if (unlikely(tc == root_task_context)) {
		root_task_context = NULL;
	} else {
		module_put(THIS_MODULE);
	}
}

static struct kobj_type task_context_type = {
	.release task_context_release_kobj,
};


int create_root_task_context(void)
{
	int err;
	task_context_t *new;
	
	if (root_task_context)
		return -EEXIST;
	
	new = kmalloc(sizeof(*new), GFP_KERNEL);
	if (!new)
		return -ENOMEM;
	memset(new, 0, sizeof(*new));

	kobject_set_name(&new->kobj, "tcontext");
	
	new->kobj.ktype = &task_context_type;
	
	err = kobject_register(&new->kobj);
	if (err) {
		kfree(new);
		return err;
	}

	root_task_context = new;
	return 0;
}
	
task_context_t *create_task_context(task_context_t *parent)
{
	int err;
	task_context_t *new = NULL;
	if (!parent)
		parent = root_task_context;
	
	if (!try_module_get(THIS_MODULE))
		return NULL;
	
	new = kmalloc(sizeof(*new), GFP_KERNEL);
	if (!new)
		goto out;
	memset(new, 0, sizeof(*new));

	kobject_set_name(&new->kobj, "%u", (unsigned int) new);
	new->kobj.ktype = &task_context_type;
	new->kobj.parent = &parent->kobj;
	
	err = kobject_register(&new->kobj);
	if (err) {
		goto out;
	}
	
	/* kobject_register sets &new->kobj.refcount to 2, so we decrement */
	/* FIXME: is this correct? */
	kobject_put(&new->kobj);
	
	return new;

out:
	if (new)
		kfree(new);
	module_put(THIS_MODULE);
	return NULL;
}

int is_task_subcontext(task_context_t *child, task_context_t *parent)
{
	if (!parent || parent == root_task_context)
		return 1;
	if (!child || child == root_task_context)
		return 0;
	do {
		if (child == parent)
			return 1;
		child = task_context_parent(child);
	} while (child != root_task_context);
	return 0;
}

int attach_task_context(struct task_struct *p, task_context_t *tc)
{
	int err;
	struct kobject *ko, *old_ko = NULL;
	if (!tc || IS_ERR(tc))
		return -EINVAL;
	if (!task_security(p)) {
		err = alloc_task_security(p);
		if (err)
			return err;
	}
	/* Get a reference to tc */
	ko = kobject_get(&tc->kobj);
	if (!ko)
		return -EBUSY;
	
	if (task_context(p))
		old_ko = &task_context(p)->kobj;
	
	task_security(p)->context = tc;
	
	/* Release the reference to the previous context, if applicable. */
	if (old_ko) {
		kobject_put(old_ko);
	}

	return 0;
}

void detach_task_context(struct task_struct *p)
{
	struct kobject *ko;
	if (!p || !task_context(p))
		return;
	ko = &task_context(p)->kobj; 
	task_security(p)->context = NULL;
	kobject_put(ko);
}

